
#import "HelperClass.h"
#import "StatusPackageModel.h"
#if TARGET_OS_IOS || TARGET_OS_TV
#import "NSTask.h"
#endif
@implementation HelperClass

/*
 tar cpJ -v -p --exclude .DS_Store --exclude docs --exclude share/doc --exclude share/info --exclude share/man --exclude man --exclude share/readline -f bs.tar.xz fs private var
 */

+ (NSArray *)excludeArgs {
    return @[@".DS_Store", @"docs", @"share/doc", @"share/info", @"share/man", @"man", @"share/readline"];
}

+ (NSInteger)createBootstrapTar:(NSString *)tarFile withBootstrap:(NSString *)bootstrap {

    __block NSMutableArray *excludeArgs = [NSMutableArray new];
    [excludeArgs addObject:@"--exclude"];
    [[self excludeArgs] enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        [excludeArgs addObject:obj];
        DLog(@"idx: %lu count: %lu", idx, excludeArgs.count);
        if (idx < excludeArgs.count-1){
            [excludeArgs addObject:@"--exclude"];
        }
    }];
    return 0;
}

+ (int)createNewBootstrap:(NSString *)bootstrapPath withPackages:(NSString *)packagesFolder {
    [FM createDirectoryAtPath:bootstrapPath withIntermediateDirectories:true attributes:nil error:nil];
    NSArray *folders = @[@"var/lib/dpkg/info",
                         @"fs/jb/usr"];
    [folders enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *path = [bootstrapPath stringByAppendingPathComponent:obj];
        [FM createDirectoryAtPath:path withIntermediateDirectories:true attributes:nil error:nil];
    }];
    NSString *statusFile = [bootstrapPath stringByAppendingPathComponent:@"var/lib/dpkg/status"];
    if (![FM fileExistsAtPath:statusFile]){
        [@"" writeToFile:statusFile atomically:true encoding:NSUTF8StringEncoding error:nil];
    }
    NSArray *packagesArray = [FM contentsOfDirectoryAtPath:packagesFolder error:nil];
    [packagesArray enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *debFile = [packagesFolder stringByAppendingPathComponent:obj];
        DLog(@"processing package: %@", debFile);
        InputPackage *output = [HelperClass packageForDeb:debFile];
        output.forceOverwrite = true;
        [output installToBootstrapPath:bootstrapPath];
    }];
    return 0;
}

+ (int)deletePackage:(NSString *)deletePackage inBootstrap:(NSString *)bootstrapPath {
    DLog(@"deleting package: %@ from bootstrap: %@", deletePackage, bootstrapPath);
    __block NSString *postRmFile = nil;
    NSString *statusFile = [bootstrapPath relativeStatusFilePath];
    NSArray <StatusPackageModel*> *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile];
    StatusPackageModel *model = [[installedPackages filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"package == %@", deletePackage]] lastObject];
    if (model) {
        NSString *listName = [NSString stringWithFormat:@"%@.list", deletePackage];
        NSString *infoPath = [bootstrapPath relativeInfoFolderPath];
        NSString *listFile = [infoPath stringByAppendingPathComponent:listName];
        NSArray *listFiles = [FM contentsOfDirectoryAtPath:infoPath error:nil];
        NSPredicate *blockPred = [NSPredicate predicateWithBlock:^BOOL(id  _Nullable evaluatedObject, NSDictionary<NSString *,id> * _Nullable bindings) {
            if ([evaluatedObject containsString:deletePackage]){
                return TRUE;
            }
            return FALSE;
        }];
        NSArray *relatedFiles = [listFiles filteredArrayUsingPredicate:blockPred]; //any files in dpkg/info/deletePackage.*
        DLog(@"\nWe have found the package, determining list details from: %@", listFile);
        NSString *listContents = [NSString stringWithContentsOfFile:listFile encoding:NSUTF8StringEncoding error:nil];
        DLog(@"\nContents: %@\n", listContents);
        DLog(@"\nFound package info files: %@\n", relatedFiles);
        NSArray *files = [listContents componentsSeparatedByString:@"\n"];
        __block NSMutableArray *filesToProcess = [NSMutableArray new];
        [relatedFiles enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            NSString *fullPath = [infoPath stringByAppendingPathComponent:obj];
            if([FM fileExistsAtPath:fullPath]){
                if ([[obj pathExtension] isEqualToString:@"postrm"]){
                    postRmFile = fullPath;
                }
                [filesToProcess addObject:fullPath];
            }
        }];
        [files enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            NSString *fullPath = [bootstrapPath stringByAppendingPathComponent:obj];
            BOOL isDir = FALSE;
            if([FM fileExistsAtPath:fullPath isDirectory:&isDir]){
                if (!isDir){
                    DLog(@"found file: %@", fullPath);
                    [filesToProcess addObject:fullPath];
                }
            }
        }];
        DLog(@"\nFiles to process: %@\n", filesToProcess);
        NSString *error = [NSString stringWithFormat:@" [WARNING] We are about to delete '%@' from the bootstrap, this cannot be undone!", deletePackage];
        if(![HelperClass shouldContinueWithPrompt:error]){
            return -1;
        }
        [filesToProcess enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            [FM removeItemAtPath:obj error:nil];
        }];
        
        NSString *statusFile = [bootstrapPath relativeStatusFilePath];
        [FM copyItemAtPath:statusFile toPath:[statusFile stringByAppendingPathExtension:@"bak"] error:nil];
        __block NSMutableString *newStatusFile = [NSMutableString new];
        [installedPackages enumerateObjectsUsingBlock:^(StatusPackageModel * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if (obj != model){
                [newStatusFile appendFormat:@"%@\n\n", obj.rawString];
            } else {
                DLog(@"skipping the model we want to delete: %@", obj);
            }
        }];
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
        [newStatusFile writeToFile:statusFile atomically:TRUE];
#pragma clang diagnostic pop
        if (postRmFile) {
            DLog(@"\n [WARNING] We found a postrm file, will not run this due to potential unexpected consequences off of the target device, displaying contents so you can determine if any additional steps are necessary\n");
            DLog(@"\n\n%@ contents:\n%@\n\n", postRmFile, [NSString stringWithContentsOfFile:postRmFile encoding:NSUTF8StringEncoding error:nil]);
        }
    }
    return 0;
}

+ (int)cleanBootstrapAtPath:(NSString *)bootstrapPath {
    NSString *statusFile = [bootstrapPath relativeStatusFilePath];
    NSString *listContents = [NSString stringWithContentsOfFile:statusFile encoding:NSUTF8StringEncoding error:nil];
    //clean up any errant spaces
    listContents = [listContents stringByReplacingOccurrencesOfString:@"\n\n\n" withString:@"\n\n"];
    listContents = [listContents stringByReplacingOccurrencesOfString:@"\n\n\n\n" withString:@"\n\n"];
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
    [listContents writeToFile:statusFile atomically:TRUE];
#pragma clang diagnostic pop
    DLog(@"\n [INFO] Cleaning status file: %@", statusFile);
    
    NSString *testStatusFile = [bootstrapPath relativeStatusFilePath];
    NSArray <StatusPackageModel*> *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile];
    __block NSMutableString *newStatusFile = [NSMutableString new];
    __block NSMutableArray *alreadyAdded = [NSMutableArray new];
    
    [installedPackages enumerateObjectsUsingBlock:^(StatusPackageModel * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        DLog(@"\n [INFO] Processing package: %@", obj.package);
        if (![alreadyAdded containsObject:obj.package]){
            //duplicate check first
            NSArray *doubleCheck = [installedPackages filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"package == %@", obj.package]];
            if (doubleCheck.count > 1){
                NSSortDescriptor *sortDesc = [NSSortDescriptor sortDescriptorWithKey:@"version" ascending:FALSE];
                StatusPackageModel *latestVersion = [[doubleCheck sortedArrayUsingDescriptors:@[sortDesc]] firstObject];
                DLog(@"\n [INFO] We found a duplicate package, choosing latestVersion: %@", latestVersion);
                [newStatusFile appendFormat:@"%@\n\n", latestVersion.rawString];
                [alreadyAdded addObject:obj.package];
            } else {
                [newStatusFile appendFormat:@"%@\n\n", obj.rawString];
            }
        } else {
            DLog(@"\n [INFO] We already had this package, skipping older duplicate: %@\n", obj.package);
        }
    }];
    testStatusFile = [testStatusFile stringByReplacingOccurrencesOfString:@"\n\n\n" withString:@"\n\n"];
    DLog(@"\n Writing new output file: %@\n", testStatusFile);
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
    [newStatusFile writeToFile:testStatusFile atomically:TRUE];
#pragma clang diagnostic pop
    DLog(@"\n Done!\n\n");
    return 0;
    
}

+ (int)repackagePackageAtPath:(NSString *)debFile withPrefix:(NSString *)prefix sigChecks:(BOOL)skip arch:(NSString *)arch excludeFile:(NSString *)excludeFile {
    DLog(@"\n [INFO] Repackaging file: %@...\n", debFile);
    InputPackage *output = [InputPackage packageForDeb:debFile excludes:excludeFile];
    output.appendedPrefix = prefix;
    output.skipSignatureChecks = skip;
    output.excludeFile = excludeFile;
    if (!arch) {
        arch = @"appletvos-arm64";
    }
    [output repackageInCurrentDirectoryWithArch:arch];
    return 0;
}

+ (int)bumpPackageAtPath:(NSString *)debFile withPrefix:(NSString *)prefix sigChecks:(BOOL)skip excludeFile:(NSString *)excludeFile {
    DLog(@"\n [INFO] Bumping version number for file: %@...\n", debFile);
    InputPackage *output = [InputPackage packageForDeb:debFile excludes:excludeFile];
    output.appendedPrefix = prefix;
    output.skipSignatureChecks = skip;
    [output bumpVersionInCurrentDirectory];
    return 0;
}

+ (NSArray <StatusPackageModel*>*)statusInstalledPackagesFromFile:(NSString *)statusFile  {
    return [self statusInstalledPackagesFromFile:statusFile bootstrapPath:nil];
}

+ (NSString *)numberedArrayFromStatusFile:(NSString *)statusFile bootstrapPath:(NSString *)boostrapPath {
    if (![FM fileExistsAtPath:statusFile]) {
        return nil;
    }
    NSArray <StatusPackageModel *> *packages = [self statusInstalledPackagesFromFile:statusFile bootstrapPath:boostrapPath];
    __block NSMutableArray *newItems = [NSMutableArray new];
    [packages enumerateObjectsUsingBlock:^(StatusPackageModel * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *currentString = [NSString stringWithFormat:@"%02lu. %@", idx+1, obj.package];
        [newItems addObject:currentString];
    }];
    return [newItems componentsJoinedByString:@"\n"];
}

+ (NSArray <StatusPackageModel*>*)statusInstalledPackagesFromFile:(NSString *)statusFile bootstrapPath:(NSString *)bootstrapPath {
    
    if (![FM fileExistsAtPath:statusFile]) {
        return nil;
    }
    NSString *packageString = [NSString stringWithContentsOfFile:statusFile encoding:NSUTF8StringEncoding error:nil];
    NSArray *lineArray = [packageString componentsSeparatedByString:@"\n\n"];
    //DLog(@"lineArray: %@", lineArray);
    NSMutableArray *mutableList = [[NSMutableArray alloc] init];
    //NSMutableDictionary *mutableDict = [[NSMutableDictionary alloc] init];
    for (id currentItem in lineArray) {
        StatusPackageModel *debModel = [[StatusPackageModel alloc] initWithRawControlString:currentItem];
        if (debModel != nil) {
            if (bootstrapPath) {
                debModel.bootstrapPath = bootstrapPath;
            }
            [mutableList addObject:debModel];
        }
    }
    NSSortDescriptor *nameDescriptor = [[NSSortDescriptor alloc] initWithKey:@"name" ascending:YES
                                                                    selector:@selector(localizedCaseInsensitiveCompare:)];
    NSSortDescriptor *packageDescriptor = [[NSSortDescriptor alloc] initWithKey:@"package" ascending:YES
                                                                       selector:@selector(localizedCaseInsensitiveCompare:)];
    NSArray *descriptors = [NSArray arrayWithObjects:packageDescriptor, nameDescriptor, nil];
    NSArray *sortedArray = [mutableList sortedArrayUsingDescriptors:descriptors];
    mutableList = nil;
    return sortedArray;
}

+ (NSInteger)getNumberFromPrompt:(NSString *)prompt {
    char c;
    char buf[64];
    int i = 0;
    printf("%s", [prompt UTF8String]);
    c=getchar();
    while (c != '\n') {
        buf[i] = c;
        i++;
        c=getchar();
    }
    buf[i] = '\0';
    return atoi(buf);
}

+ (BOOL)shouldContinueWithPrompt:(NSString *)prompt {
    NSString *promptString = [NSString stringWithFormat:@"\n%@ Are you sure you want to continue? [y/n]?", prompt];
    char c;
    printf("%s", [promptString UTF8String] );
    c=getchar();
    while(c!='y' && c!='n') {
        if (c!='\n'){
            printf("[y/n]");
        }
        c=getchar();
    }
    if (c == 'n') {
        DLog(@"\nSmart move... exiting\n\n");
        return FALSE;
    } else if (c == 'y') {
        DLog(@"\nDon't say we didn't warn ya!....\n");
    }
    return TRUE;
}

+ (NSArray *)arrayReturnForTask:(NSString *)taskBinary withArguments:(NSArray *)taskArguments {
    DLog(@"%@ %@", taskBinary, [taskArguments componentsJoinedByString:@" "]);
    NSTask *task = [[NSTask alloc] init];
    NSPipe *pipe = [[NSPipe alloc] init];
    NSFileHandle *handle = [pipe fileHandleForReading];
    [task setLaunchPath:taskBinary];
    [task setArguments:taskArguments];
    [task setStandardOutput:pipe];
    [task setStandardError:pipe];
    [task launch];
    NSData *outData = nil;
    NSString *temp = nil;
    while((outData = [handle readDataToEndOfFile]) && [outData length]) {
        temp = [[NSString alloc] initWithData:outData encoding:NSASCIIStringEncoding];
    }
    [handle closeFile];
    task = nil;
    return [temp componentsSeparatedByString:@"\n"];
}

+ (NSString *)singleLineReturnForProcess:(NSString *)format, ... {
    if (format==nil)
        return nil;
    char line[200];
    va_list args;
    va_start(args, format);
    NSString *newString = [[NSString alloc] initWithFormat:format arguments:args];
    va_end(args);
    //DLog(@"running process: %@", newString);
    FILE* fp = popen([newString UTF8String], "r");
    NSMutableString *lines = [[NSMutableString alloc]init];
    if (fp) {
        while (fgets(line, sizeof line, fp)) {
            NSString *s = [NSString stringWithCString:line encoding:NSUTF8StringEncoding];
            s = [s stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
            [lines appendString:s];
        }
    }
    pclose(fp);
    return lines;
}

+ (NSArray *)returnForProcess:(NSString *)call {
    if (call==nil)
        return 0;
    char line[200];
    //DLog(@"\nRunning process: %@\n", call);
    FILE* fp = popen([call UTF8String], "r");
    NSMutableArray *lines = [[NSMutableArray alloc]init];
    if (fp) {
        while (fgets(line, sizeof line, fp)) {
            NSString *s = [NSString stringWithCString:line encoding:NSUTF8StringEncoding];
            s = [s stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
            [lines addObject:s];
        }
    }
    pclose(fp);
    return lines;
}

+ (int)runCommand:(NSString *)call environment:(NSDictionary *)env currentPath:(NSString *)currentPath verbose:(BOOL)verbose {
    if (call==nil)
        return 0;
    char line[200];
    chdir([currentPath UTF8String]);
    FILE* fp = popen([call UTF8String], "r");
    //NSMutableArray *lines = [[NSMutableArray alloc]init];
    if (fp) {
        while (fgets(line, sizeof line, fp)) {
            if (verbose) {
                NSString *s = [NSString stringWithCString:line encoding:NSUTF8StringEncoding];
                s = [s stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
                DLog(@"%@",s);
            }
        }
    }
    int returnCode = pclose(fp);
    return returnCode;
}

+ (void)runProcess:(NSString *)call environment:(NSDictionary *)env currentPath:(NSString *)currentPath completion:(void(^)(NSString *output, NSInteger returnStatus))block {

    NSArray *args = [call componentsSeparatedByString:@" "];
    NSString *taskBinary = args[0];
    NSArray *taskArguments = [args subarrayWithRange:NSMakeRange(1, args.count-1)];
    DLog(@"%@ %@", taskBinary, [taskArguments componentsJoinedByString:@" "]);
    NSTask *task = [[NSTask alloc] init];
    NSPipe *pipe = [[NSPipe alloc] init];
    NSFileHandle *handle = [pipe fileHandleForReading];
    [task setLaunchPath:taskBinary];
    [task setArguments:taskArguments];
    [task setEnvironment:env];
    [task setCurrentDirectoryPath:currentPath];
    [task setStandardOutput:pipe];
    [task setStandardError:pipe];
    [task launch];
    NSData *outData = nil;
    NSString *temp = nil;
    while((outData = [handle readDataToEndOfFile]) && [outData length]){
        temp = [[NSString alloc] initWithData:outData encoding:NSASCIIStringEncoding];
        DLog(@"%@", temp);
    }
    [handle closeFile];
    [task waitUntilExit];
    int termStatus = [task terminationStatus];
    task = nil;
    if (block){
        block(temp, termStatus);
    }
}
//-rw-r--r--   1 root        wheel        4197184 May  9 19:53 nitotv_3.4.3-63_appletvos-arm64.deb
+ (InputPackageFile *)packageFileFromLine:(NSString *)inputLine {
    //    "-rwxr-xr-x  0 root   wheel   69424 Oct 22 03:56 ./Library/MobileSubstrate/DynamicLibraries/beigelist7.dylib\n",
    
    //-rwxr-xr-x root/staff    10860 2011-02-02 03:55 ./Library/Frameworks/CydiaSubstrate.framework/Commands/cycc
    NSArray *lineObjects = [inputLine spaceDelimitedArray];
    
    //DLog(@"lineObjects: %@", lineObjects);
    /*
     
     "drwxr-xr-x",
     "root/wheel",
     0,
     "2018-06-27",
     "01:21",
     "./"
     */
    NSString *permissionsAndType = [lineObjects objectAtIndex:0];
    NSString *userGroup = [lineObjects objectAtIndex:1];
    NSString *size = [lineObjects objectAtIndex:2];
    NSString *date = [lineObjects objectAtIndex:3];
    NSString *time = [lineObjects objectAtIndex:4];
    NSString *path = [lineObjects objectAtIndex:5];
    
    //@"drwxr-xr-x"
    NSString *fileTypeChar = [permissionsAndType substringWithRange:NSMakeRange(0, 1)];
    NSString *octalPermissions = [self octalFromSymbols:permissionsAndType];
    NSString *octalUG = [self octalFromGroupSymbols:userGroup];
    NSString *fileName = [path lastPathComponent];
    //NSString *fullPath = [NSString stringWithFormat:@"/%@", path];
    NSString *fullPath = path;
    if ([path hasPrefix:@"."]){
        fullPath = [path substringFromIndex:1];
    }
    
    InputPackageFile *pf = [InputPackageFile new];
    [pf _setFileTypeFromRaw:fileTypeChar];
    
    switch (pf.type) {
        case BSPackageFileTypeLink:
        {
            fullPath = [lineObjects objectAtIndex:7];
            NSString *linkDest = [NSString stringWithFormat:@"/%@", path];
            pf.permissions = octalPermissions;
            pf.owner = octalUG;
            pf.size = size;
            pf.time = time;
            pf.date = date;
            pf.path = fullPath;
            pf.basename = fileName;
            pf.linkDestination = linkDest;
            
            return pf;
        }
            break;
            
        case BSPackageFileTypeDirectory:
            //DLog(@"we dont want directory entries do we %@", lineObjects);
            pf.permissions = octalPermissions;
            pf.owner = octalUG;
            pf.size = size;
            pf.time = time;
            pf.date = date;
            pf.path = fullPath;
            pf.basename = fileName;
            return pf;
            break;
            
        default:
            break;
    }
    pf.permissions = octalPermissions;
    pf.owner = octalUG;
    pf.size = size;
    pf.time = time;
    pf.date = date;
    pf.path = fullPath;
    pf.basename = fileName;
    return pf;
}

+ (NSString *)octalFromGroupSymbols:(NSString *)theSymbols {
    NSArray *groupArray = [theSymbols componentsSeparatedByString:@"/"];
    NSString *user = [groupArray objectAtIndex:0];
    NSString *group = [groupArray objectAtIndex:1];
    NSString *octalUser = nil;
    NSString *octalGroup = nil;
    //uid=0(root) gid=0(wheel) groups=0(wheel),1(daemon),2(kmem),3(sys),4(tty),5(operator),8(procview),9(procmod),20(staff),29(certusers),80(admin)
    if ([user isEqualToString:@"root"]) {
        octalUser = @"0";
    } else if ([user isEqualToString:@"mobile"]) {
        octalUser = @"501";
    }
    //obviously more cases!! FIXME:
    if ([group isEqualToString:@"staff"]) {
        octalGroup = @"20";
    } else if ([group isEqualToString:@"admin"]) {
        octalGroup = @"80";
    } else if ([group isEqualToString:@"wheel"]) {
        octalGroup = @"0";
    } else if ([group isEqualToString:@"daemon"]) {
        octalGroup = @"1";
    } else if ([group isEqualToString:@"kmem"]) {
        octalGroup = @"2";
    } else if ([group isEqualToString:@"sys"]) {
        octalGroup = @"3";
    } else if ([group isEqualToString:@"tty"]) {
        octalGroup = @"4";
    } else if ([group isEqualToString:@"operator"]) {
        octalGroup = @"5";
    } else if ([group isEqualToString:@"procview"]) {
        octalGroup = @"8";
    } else if ([group isEqualToString:@"procmod"]) {
        octalGroup = @"9";
    } else if ([group isEqualToString:@"certusers"]) {
        octalGroup = @"29";
    } else {
        octalGroup = @"501"; //default to mobile
    }
    //uid=0(root) gid=0(wheel) groups=0(wheel),1(daemon),2(kmem),3(sys),4(tty),5(operator),8(procview),9(procmod),20(staff),29(certusers),80(admin)
    return [NSString stringWithFormat:@"%@:%@", octalUser, octalGroup];
}

+ (NSString *)validLDIDPath {
    NSString *ldid = [[HelperClass arrayReturnForTask:@"/usr/bin/which" withArguments:@[@"ldid"]] firstObject];
    if (ldid) {
        //DLog(@"found ldid: %@", ldid);
        NSString *ldidCommand = [NSString stringWithFormat:@"%@ -- 2>&1", ldid];
        NSString *ldidTest = [HelperClass singleLineReturnForProcess:ldidCommand];
        //DLog(@"ldidTest: %@", ldidTest);
        if ([ldidTest containsString:@"Link Identity Editor"]) {
            return ldid;
        }
    }
    return false;
}

+ (InputPackage *)packageForDeb:(NSString *)debFile {
    
    NSString *dpkg = [@"dpkg" runPathForSearchPath];
    NSString *packageName = [self singleLineReturnForProcess:@"%@ -f %@ Package",dpkg, debFile];
    NSString *packageVersion = [self singleLineReturnForProcess:@"%@ -f %@ Version", dpkg, debFile];
    NSArray <InputPackageFile *> *fileList = [self returnForProcess:[NSString stringWithFormat:@"%@ -c %@", dpkg, debFile]];
    
    __block NSMutableArray *finalArray = [NSMutableArray new];
    
    [fileList enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        
        InputPackageFile *file = [self packageFileFromLine:obj];
        if (file) {
            //DLog(@"%@", file);
            [finalArray addObject:file];
        }
        
    }];
    
    InputPackage *pkg = [InputPackage new];
    pkg.files = finalArray;
    pkg.path = debFile;
    pkg.packageName = packageName;
    pkg.version = packageVersion;
    pkg.validLDIDPath = [self validLDIDPath];
    return pkg;
    
}

+ (NSString *)octalFromSymbols:(NSString *)theSymbols {
    //DLog(@"%@ %s", self, _cmd);
    NSString *U = [theSymbols substringWithRange:NSMakeRange(1, 3)];
    NSString *G = [theSymbols substringWithRange:NSMakeRange(4, 3)];
    NSString *O = [theSymbols substringWithRange:NSMakeRange(7, 3)];
    //DLog(@"fileTypeChar: %@", fileTypeChar);
    //DLog(@"U; %@", U);
    //DLog(@"G; %@", G);
    //DLog(@"O; %@", O);
    
    //USER
    
    int sIdBit = 0;
    int uOctal = 0;
    const char *uArray = [U cStringUsingEncoding:NSASCIIStringEncoding];
    NSInteger stringLength = [U length];
    
    int x;
    for( x=0; x<stringLength; x++ ) {
        unsigned int aCharacter = uArray[x];
        if (aCharacter == 'r') {
            uOctal += 4;
        } else if (aCharacter == 'w') {
            uOctal += 2;
        } else if (aCharacter == 'x') {
            uOctal += 1;
        } else if (aCharacter == 's') {
            sIdBit += 4;
        }
    }
    
    //GROUP
    
    int gOctal = 0;
    const char *gArray = [G cStringUsingEncoding:NSASCIIStringEncoding];
    stringLength = [G length];
    
    int y;
    for( y=0; y<stringLength; y++ ) {
        unsigned int aCharacter = gArray[y];
        if (aCharacter == 'r') {
            gOctal += 4;
        } else if (aCharacter == 'w') {
            gOctal += 2;
        } else if (aCharacter == 'x') {
            gOctal += 1;
        } else if (aCharacter == 's') {
            gOctal += 2;
        }
    }
    //OTHERS
    int z;
    int oOctal = 0;
    const char *oArray = [O cStringUsingEncoding:NSASCIIStringEncoding];
    stringLength = [O length];
    for( z=0; z<stringLength; z++ ) {
        unsigned int aCharacter = oArray[z];
        if (aCharacter == 'r') {
            oOctal += 4;
        } else if (aCharacter == 'w') {
            oOctal += 2;
        } else if (aCharacter == 'x') {
            oOctal += 1;
        }
    }
    return [NSString stringWithFormat:@"%i%i%i%i", sIdBit, uOctal, gOctal, oOctal];
}

@end
