#import "InputPackage.h"
#import "ErrorReturn.h"
#import "HelperClass.h"
#import "Universalize.h"
#import "InputPackageFile.h"

@implementation InputPackage

+ (InputPackage *)packageForDeb:(NSString *)debFile {
    return [self packageForDeb:debFile excludes:nil];
}

+ (InputPackage *)packageForDeb:(NSString *)debFile excludes:(NSString *)excludesFile {
    
    NSString *dpkg = [@"dpkg" runPathForSearchPath];
    NSString *packageName = [HelperClass singleLineReturnForProcess:@"%@ -f %@ Package",dpkg, debFile];
    NSString *packageVersion = [HelperClass singleLineReturnForProcess:@"%@ -f %@ Version", dpkg, debFile];
    NSString *architecture = [HelperClass singleLineReturnForProcess:@"%@ -f %@ Architecture", dpkg, debFile];
    NSArray <InputPackageFile *> *fileList = [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -c %@", dpkg, debFile]];
    
    __block NSMutableArray *finalArray = [NSMutableArray new];
    
    [fileList enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        InputPackageFile *file = [HelperClass packageFileFromLine:obj];
        if (file) {
            //DLog(@"%@", file);
            [finalArray addObject:file];
        }
    }];
    
    InputPackage *pkg = [InputPackage new];
    pkg.excludeFile = excludesFile;
    pkg.files = finalArray;
    pkg.path = debFile;
    pkg.packageName = packageName;
    pkg.version = packageVersion;
    pkg.architecture = architecture;
    pkg.validLDIDPath = [HelperClass validLDIDPath];
    return pkg;
    
}

- (NSString*) description {
    
    NSString *orig = [super description];
    NSMutableDictionary *details = [NSMutableDictionary new];
    NSArray *props = [self properties];
    [props enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        
        NSString *cv = [self valueForKey:obj];
        if (cv){
            details[obj] = cv;
        }
        
    }];
    return [NSString stringWithFormat:@"%@ = %@", orig, details];
    
}

- (NSArray *)rootPaths {
    NSMutableArray *newArray = [NSMutableArray new];
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSArray *pathC = [obj.path properPathComponents];
        //DLog(@"%@ count: %lu", obj.path, pathC.count);
        if (pathC.count == 1 && ![newArray containsObject:pathC.firstObject]) {
            [newArray addObject:pathC.firstObject];
        }
    }];
    return newArray;
}

- (NSString *)listfile {
    
    __block NSMutableArray *outFiles = [NSMutableArray new];
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        
        if ([obj.fileType isEqualToString:@"link"]){ //does this need to handle things differently?
            [outFiles addObject:obj.linkDestination];
        } else {
            [outFiles addObject:obj.path];
        }
    }];
    return [outFiles componentsJoinedByString:@"\n"];
    
}

- (BOOL)validateFileList:(NSArray *)fileList {
    __block BOOL retv = true;
    __block int invalidCount = 0; //make sure its greater than 1 before we bail...
    [fileList enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        if ([obj length] > 1) {
            NSString *sub = [obj substringToIndex:2];
            if ([sub isEqualToString:@"./"]) {
                invalidCount++;
                if (invalidCount > 1){
                    *stop = true;
                    retv = false;
                }
            }
        }
    }];
    return retv;
}

- (NSArray *)getFileListing {
    NSString *dpkgDeb = [@"dpkg-deb" runPathForSearchPath];
    NSString *listProcess = [NSString stringWithFormat:@"%@ -c %@ | awk '{ print $6 }'" ,dpkgDeb, self.path];
    NSArray *fileList = [HelperClass returnForProcess:listProcess];
    BOOL validated = [self validateFileList:fileList];
    if (!validated) {
        fileList = [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -c %@ | awk '{ print $6 }' | cut -c 2-" ,dpkgDeb, self.path]];
    }
    if ([fileList containsObject:@"/"]) {
        NSMutableArray *_mutableList = [fileList mutableCopy];
        [_mutableList removeObject:@"/"];
        fileList = [_mutableList copy];
    }
    NSArray *excludeArray = [self.excludeFile excludeArrayFromFile];
    if (excludeArray) {
        __block NSMutableArray *filesToRemove = [NSMutableArray new];
        [excludeArray enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if ([obj length] > 0){
                NSArray *items = [fileList filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"SELF LIKE[c] %@", obj]];
                if (items.count > 0) {
                    //DLog(@"found items to remove: %@", items);
                    [filesToRemove addObjectsFromArray:items];
                }
            }
        }];
        //DLog(@"\nShould remove the following: %@", filesToRemove);
        self.excludedFiles = filesToRemove;
        NSMutableArray *_mutableList = [fileList mutableCopy];
        [_mutableList removeObjectsInArray:filesToRemove];
        fileList = [_mutableList copy];
    }
    return fileList;
}

- (NSString *)fullPackageName {
    return [NSString stringWithFormat:@"%@_%@_%@.deb", self.packageName, self.version, self.architecture];
}

- (int)installToBootstrapPath:(NSString *)bootstrapPath {
    
    DLog(@"\nFound package: '%@' at version: '%@'...\n", self.packageName, self.version );
    
    NSString *statusFile = [bootstrapPath relativeStatusFilePath];
    NSArray *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile];
    NSString *dpkg = [@"dpkg" runPathForSearchPath];
    //DLog(@"installedPackages: %@", installedPackages);
    
    
    StatusPackageModel *model = [[installedPackages filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"package == %@", self.packageName]] lastObject];
    
    if (model) {
        
        DLog(@"\n [WARNING] This package has already been installed! %@", model);
        NSComparisonResult result = [self.version compare:model.version options:NSNumericSearch];
        DLog(@"\n [WARNING] Comparing version numbers...");
        switch (result) {
            case NSOrderedSame:
                DLog(@"\n [WARNING] %@ and %@ match!", self.version, model.version);
                break;
                
            case NSOrderedAscending:
                DLog(@"\n [WARNING] Package version %@ is less than installed version %@!", self.version, model.version);
                break;
                
            case NSOrderedDescending:
                DLog(@"\n [WARNING] Package version %@ is greater than installed version %@!", self.version, model.version);
                break;
                
            default:
                break;
        }
        
        
        
    }
    
    ErrorReturn *safePackage = [self errorReturnForBootstrap:bootstrapPath];
    
    if (safePackage.returnStatus != 0) { //zero is success
        
        if (safePackage.returnStatus == 1) //ovewrwrites, just warnings!
        {
            
            NSString *error = [NSString stringWithFormat:@" [WARNING] %@ will overwrite the following files: \n\n\t%@\n\n", self.path.lastPathComponent, [safePackage.overwriteFiles componentsJoinedByString:@"\n\t"]];
            if (!self.forceOverwrite) {
                if(![HelperClass shouldContinueWithPrompt:error]){
                    
                    return -1;
                }
            }
        } else if (safePackage.returnStatus == 2) //bail!!"
        {
            if (![self forceInstall]) {
                return 2;
            }
        }
    }
    
    NSString *pwd = [[HelperClass returnForProcess:@"/bin/pwd"] componentsJoinedByString:@"\n"];
    
    NSString *tmpPath = [pwd stringByAppendingPathComponent:self.packageName];
    NSString *debian = [tmpPath stringByAppendingPathComponent:@"DEBIAN"];
    [FM createDirectoryAtPath:tmpPath withIntermediateDirectories:TRUE attributes:nil error:nil];
    DLog(@"\nExtracting deb for processing...\n");
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -x %@ %@", dpkg, self.path, tmpPath]];
    
    NSArray *fileList = [self getFileListing];
    
    if (fileList.count > 0) {
        
        NSString *baseFile = [NSString stringWithFormat:@"%@.list", self.packageName];
        
        NSString *listFile = [bootstrapPath.relativeInfoFolderPath stringByAppendingPathComponent:baseFile];
        NSString *listOutput = [fileList componentsJoinedByString:@"\n"];
        DLog(@"\nCreating list file '%@'...\n", listFile);
        //DLog(@"list: %@", listOutput);
        [[listOutput stringByAppendingString:@"\n"] writeToFile:listFile atomically:TRUE encoding:NSUTF8StringEncoding error:nil];
    }
    
    /*
     (dpkg-deb -c "$DEB" | awk "{ print $6}" | grep "\./" -c)
     
    NSString *process = [NSString stringWithFormat:@"/usr/local/bin/dpkg-deb -c %@ | awk '{ print $6 }'" , self.path];
    NSArray *psc = [HelperClass returnForProcess:process];
    DLog(@"psc: %@", psc);
    */
    //NSString *md5s = [bootstrapInfoPath stringByAppendingPathComponent:[NSString stringWithFormat:@"%@.md5sums", self.packageName]];
    
    //DLog(@"\nCreating list file '%@'...\n", listFile);
    
    //NSString *listString = [NSString stringWithFormat:@"/usr/bin/find %@ -type f -not -path \"*.DS_Store*\" | /usr/bin/sed \"s|%@||g\" >> \"%@\"", tmpPath, tmpPath, listFile];
    //NSString *listOutput = [HelperClass singleLineReturnForProcess:listString];
    
    //DLog(@"list output: %@", listOutput);
    
    //[listOutput writeToFile:listFile atomically:TRUE encoding:NSASCIIStringEncoding error:nil];
    
    //find "$BUILD_ROOT/$WORKING_DIR/deb" -type f -not -path "$BUILD_ROOT/$WORKING_DIR/deb/DEBIAN/*" | sed "s|$BUILD_ROOT/$WORKING_DIR/deb||g" >> "$SKEL_PREFIX/var/lib/dpkg/info/$PKG_NAME.list"
    
    //find "$BUILD_ROOT/$WORKING_DIR/deb" -type f -not -path "$BUILD_ROOT/$WORKING_DIR/deb/DEBIAN/*" -exec $MD5 {} \; |  awk '{ print $1 "  " $2 }' | sed "s|$BUILD_ROOT/$WORKING_DIR/deb/||g" >> "$SKEL_PREFIX/var/lib/dpkg/info/$PKG_NAME.md5sums"
    
    //[self.listfile writeToFile:listFile atomically:TRUE encoding:NSASCIIStringEncoding error:nil];
    /*
    DLog(@"\nGenerating md5sums...\n");
    NSString *runString = [NSString stringWithFormat:@"/usr/bin/find %@ -type f -not -path \"*.DS_Store*\" -exec /sbin/md5 -r {} \\; |  /usr/bin/awk '{ print $1 \"  \" $2 }' | /usr/bin/sed \"s|%@||g\" >> \"%@\"", tmpPath, tmpPath, md5s];
    NSString *outputs = [[HelperClass returnForProcess:runString] componentsJoinedByString:@"\n"];
    
    DLog(@"\nCreating md5sum file '%@'...\n", md5s);
    
    //[outputs writeToFile:md5s atomically:TRUE encoding:NSASCIIStringEncoding error:nil];
    */
    [FM createDirectoryAtPath:debian withIntermediateDirectories:TRUE attributes:nil error:nil];
    
    DLog(@"\nExtracting DEBIAN files for processing...\n");
    
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -e %@ %@", dpkg, self.path, debian]];
    //NSString *nextPath = [tmpPath stringByAppendingPathComponent:@"DEBIAN"];
    
    DLog(@"\nCopying any necessary DEBIAN files to new locations...\n\n");
    
    __block NSString *postInstFile = nil;
    
    NSArray *files = [FM contentsOfDirectoryAtPath:debian error:nil];
    [files enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        
        NSString *fullPath = [debian stringByAppendingPathComponent:obj];
        if (![obj isEqualToString:@"control"]){
            if ([obj isEqualToString:@"postinst"]){
                postInstFile = fullPath;
            }
            DLog(@"Copying file... %@\n\n", fullPath);
            NSString *fileName = [NSString stringWithFormat:@"%@.%@", self.packageName, obj];
            NSString *newPath = [bootstrapPath relativeInfoFolderPath];
            newPath = [newPath stringByAppendingPathComponent:fileName];
            //DLog(@"newPath: %@", newPath);
            [FM copyItemAtPath:fullPath toPath:newPath error:nil];
            
        } else {
            
            NSMutableString *controlFile = [[NSMutableString alloc] initWithContentsOfFile:fullPath encoding:NSASCIIStringEncoding error:nil];
            
            //[controlFile replaceOccurrencesOfString:@"iphoneos-arm" withString:@"appletvos-arm64" options:NSLiteralSearch range:NSMakeRange(0, [controlFile length])];
            
            NSString *controlString = self.unpack ? @"Status: install ok unpacked" : @"Status: install ok installed";
            
            [controlFile appendString:controlString];
            
            //DLog(@"control file: -%@-\n", controlFile);
            
            NSString *statusFile = [bootstrapPath relativeStatusFilePath];
            
            [FM copyItemAtPath:statusFile toPath:[statusFile stringByAppendingPathExtension:@"bak"] error:nil];
            
            NSMutableString *statusContents = [[NSMutableString alloc] initWithContentsOfFile:statusFile encoding:NSASCIIStringEncoding error:nil];
            
            if (model) { //we found the package model, just replace the old string with our new one
                DLog(@"Updating status file for new package version...\n\n");
                
                [statusContents replaceOccurrencesOfString:model.rawString withString:controlFile options:NSLiteralSearch range:NSMakeRange(0, [statusContents length])];
                //[statusContents writeToFileWithoutAttributes:statusFile];
                [statusContents writeToFile:statusFile atomically:TRUE encoding:NSASCIIStringEncoding error:nil];
            } else {
                DLog(@"Updating status file for new package...\n\n");
                
                //DLog(@"down here??: -%@-\n", controlFile);
                [statusContents appendFormat:@"%@\n", controlFile];
                //DLog(@"\nnew status file: %@\n", statusContents);
                //[statusContents writeToFileWithoutAttributes:statusFile];
                [statusContents writeToFile:statusFile atomically:TRUE encoding:NSASCIIStringEncoding error:nil];
            }
        }
    }];
    
    //finally actually install the package onto the bootstrap
    
    DLog(@"Extracting package onto bootstrap folder...\n");
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -x %@ %@",dpkg, self.path, bootstrapPath]];
    if (postInstFile) {
        DLog(@"\n [WARNING] We found a postinst file, will not run this due to potential unexpected consequences in your run environment! Displaying contents so you can determine if any additional steps are necessary. Contents will be delimited by a line -----\n");
        DLog(@"\n%@ Contents:\n\n---------------------\n\n%@\n\n---------------------\n\n", postInstFile, [NSString stringWithContentsOfFile:postInstFile encoding:NSASCIIStringEncoding error:nil]);
    }
    NSArray *excludeArray = self.excludedFiles;//[self.excludeFile excludeArrayFromFile];
    if (excludeArray) {
        DLog(@"\nRemoving excluded files...\n\n");
        [excludeArray enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if ([obj length] > 0){
                NSString *filePath = [bootstrapPath stringByAppendingPathComponent:obj];
                if ([FM fileExistsAtPath:filePath]) {
                    NSString *commandToRun = [NSString stringWithFormat:@"/bin/rm -r %@", filePath];
                    DLog(@"%@", commandToRun);
                    [HelperClass singleLineReturnForProcess:commandToRun];
                } else {
                    DLog(@"file doesn't exist at %@", filePath);
                }
                //DLog(@"returned: %@", deets);
            }
        }];
    }
    [FM removeItemAtPath:tmpPath error:nil];
    DLog(@"\nDone!\n\n");
    return 0;
}

- (BOOL)fileIsSymbolicLink:(NSString *)file {
    NSString *fileType = [FM attributesOfItemAtPath:file error:nil][NSFileType];
    return ([fileType isEqualToString:NSFileTypeSymbolicLink] || [fileType isEqualToString:NSFileTypeDirectory]);
}

- (void)flattenIfNecessary:(NSString *)file {
    if ([self fileIsSymbolicLink:file] || ![self shouldCodesignFile:file]){
        //DLog(@"skipping file: %@", file);
        return;
    }
    if (![FM fileExistsAtPath:file]){
        return;
    }
    DLog(@"processing file: %@", file);
    NSString *fat = [HelperClass singleLineReturnForProcess:@"/usr/bin/lipo -info %@",file];
    if ([fat containsString:@"Architectures"]){
        NSString *newFile = [file stringByAppendingPathExtension:@"thin"];
        [HelperClass singleLineReturnForProcess:[NSString stringWithFormat:@"/usr/bin/lipo -thin arm64 %@ -output %@",file, newFile]];
        if ([FM fileExistsAtPath:newFile]){
            [FM removeItemAtPath:file error:nil];
            [FM moveItemAtPath:newFile toPath:file error:nil];
        }
        /*
        int uni = [Universalize universalize:file];
        if (uni == 1){
            DLog(@"Universalized %@ successfully!", file);
            [self codesignRetainingSignature:file];
        } else {
            DLog(@"Universalize %@ failed or non zero!: %d", file, uni);
        } */
        // @"sudo lipo -thin arm64 $input.og -output $input";
    }
}

- (void)codesignRetainingSignature:(NSString *)file {
    NSString *ldidPath = [self validLDIDPath];
    if (ldidPath) {
        [FM removeItemAtPath:@"/tmp/ent.plist" error:nil];
        [HelperClass singleLineReturnForProcess:@"%@ -e %@ > /tmp/ent.plist", ldidPath, file];
        NSString *fileCheck = [NSString stringWithContentsOfFile:@"/tmp/ent.plist" encoding:NSUTF8StringEncoding error:nil];
        if (fileCheck.length == 0) {
            [HelperClass singleLineReturnForProcess:@"%@ -S %@", ldidPath, file];
        } else {
            [HelperClass singleLineReturnForProcess:@"%@ -S/tmp/ent.plist %@", ldidPath, file];
        }
    }
}

- (BOOL)shouldCodesignFile:(NSString *)file {
    NSArray *codesignPrefixes = @[@"py", @"so", @"dylib", @"appex"];
    BOOL isDir = false;
    BOOL exists = [FM fileExistsAtPath:file isDirectory:&isDir];
    if (isDir || !exists) return false;
    return ([FM isExecutableFileAtPath:file] || [codesignPrefixes containsObject:[[file pathExtension] lowercaseString]]);
}

- (void)codesignIfNecessary:(NSString *)file {
    if ([self shouldCodesignFile:file]){
        [self codesignRetainingSignature:file];
    }
}

- (void)flattenInPath:(NSString *)thePath {
    if (self.skipSignatureChecks) { return; }
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *file = [thePath stringByAppendingPathComponent:obj.path];
        //DLog(@"check sig file: %@", obj.path);
        [self flattenIfNecessary:file];
    }];
    
}

- (void)validateSignaturesInPath:(NSString *)thePath {
    if (self.skipSignatureChecks) { return; }
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *file = [thePath stringByAppendingPathComponent:obj.path];
        //DLog(@"check sig file: %@", file);
        [self codesignIfNecessary:file];
    }];
    
}

- (void)sanitizeScriptsIfNecessary:(NSString *)debianFolder {
    if (self.appendedPrefix.length > 0) {
        NSArray *files = [FM contentsOfDirectoryAtPath:debianFolder error:nil];
        [files enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if (![obj isEqualToString:@"control"]) {
                NSString *fullPath = [debianFolder stringByAppendingPathComponent:obj];
                DLog(@"checking file: %@", fullPath);
                NSString *fullString = [NSString stringWithContentsOfFile:fullPath encoding:NSUTF8StringEncoding error:nil];
                if (!fullString) {
                    fullString = [NSString stringWithContentsOfFile:fullPath encoding:NSASCIIStringEncoding error:nil];
                }
                if ([fullString containsString:@"#!/bin/bash"] || [fullString containsString:@"#!/bin/sh"]){
                    fullString = [fullString stringByReplacingOccurrencesOfString:@"#!/bin/bash" withString:@""];
                    fullString = [fullString stringByReplacingOccurrencesOfString:@"#!/bin/sh" withString:@""];
                    if (fullString.length > 0) {
                        [fullString writeToFile:fullPath atomically:TRUE encoding:NSUTF8StringEncoding error:nil];
                        [HelperClass singleLineReturnForProcess:@"chmod +x %@", fullPath];
                    }
                }
            }
        }];
    }
}

- (void)bumpVersionInCurrentDirectory {
    
    NSString *fakeRoot = [HelperClass singleLineReturnForProcess:@"/usr/bin/which fakeroot"];
    NSString *pwd = [HelperClass singleLineReturnForProcess:@"/bin/pwd"];
    NSString *dpkg = [@"dpkg" runPathForSearchPath];
    DLog(@"\nProcessing file: %@\n", self.path);
    InputPackage *output = self;
    NSMutableArray *moveRoot = [NSMutableArray new];
    DLog(@"\nFound package: '%@' at version: '%@'...\n", output.packageName, output.version );
    
    NSString *tmpPath = [pwd stringByAppendingPathComponent:output.packageName];
    NSString *debian = [tmpPath stringByAppendingPathComponent:@"DEBIAN"];
    NSString *newRoot = nil;
    if (self.appendedPrefix) {
        newRoot = [tmpPath stringByAppendingPathComponent:self.appendedPrefix];
        [FM createDirectoryAtPath:newRoot withIntermediateDirectories:true attributes:nil error:nil];
    }
    [FM createDirectoryAtPath:tmpPath withIntermediateDirectories:TRUE attributes:nil error:nil];
    DLog(@"\nExtracting package contents for processing...\n");
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -x %@ %@",dpkg, self.path, tmpPath]];
    [FM createDirectoryAtPath:debian withIntermediateDirectories:TRUE attributes:nil error:nil];
    DLog(@"\nExtracting DEBIAN files for processing...\n");
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -e %@ %@",dpkg, self.path, debian]];
    NSArray *excludeArray = [self.excludeFile excludeArrayFromFile];
    if (excludeArray) {
        [excludeArray enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if ([obj length] > 0){
                NSString *commandToRun = [NSString stringWithFormat:@"/bin/rm %@/%@", tmpPath, obj];
                //DLog(@"%@", commandToRun);
                NSString *deets = [HelperClass singleLineReturnForProcess:commandToRun];
                DLog(@"returned: %@", deets);
            }
        }];
    }
    NSString *controlPath = [debian stringByAppendingPathComponent:@"control"];
    NSMutableString *controlFile = [[NSMutableString alloc] initWithContentsOfFile:controlPath encoding:NSASCIIStringEncoding error:nil];
    //@"appletvos-arm64"
    NSString *nextVersionNumber = [self.version nextVersionNumber];
    [controlFile replaceOccurrencesOfString:self.version withString:nextVersionNumber options:NSLiteralSearch range:NSMakeRange(0, [controlFile length])];
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
    [controlFile writeToFile:controlPath atomically:TRUE];
    self.version = nextVersionNumber;
    //[self sanitizeScriptsIfNecessary:debian];

    //at this point we have the files extracted, time to determine what needs to be changed
    
    NSArray *ignoreFiles = @[@".fauxsu", @".DS_Store"];
    NSArray *forbiddenRoots = @[@"etc", @"var", @"tmp"];
    [self flattenInPath:tmpPath];
    [self validateSignaturesInPath:tmpPath];
    
    //__block NSMutableArray *_overwriteArray = [NSMutableArray new];
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        if (obj.type == BSPackageFileTypeFile || obj.type == BSPackageFileTypeDirectory){
            //DLog(@"processing path: %@", obj.path);
            if ([obj.path isEqualToString:@"/private/var/mobile/Applications/"]){
                NSString *badPath = [tmpPath stringByAppendingPathComponent:obj.path];
                NSString *newPath = [tmpPath stringByAppendingPathComponent:@"Applications"];
                DLog(@"\n [INFO] Moving %@ to %@...", badPath, newPath);
                [FM moveItemAtPath:badPath toPath:newPath error:nil];
                [FM removeItemAtPath:[tmpPath stringByAppendingPathComponent:@"private"] error:nil];
                *stop = TRUE;
            }
            if ([[obj.path lastPathComponent] isEqualToString:@"Info.plist"]){
                if ([[[[obj.path stringByDeletingLastPathComponent] lastPathComponent] pathExtension] isEqualToString:@"app"]){
                    DLog(@"Got an Info.plist inside an app: %@", obj.path);
                    NSString *plistFullPath = [tmpPath stringByAppendingPathComponent:obj.path];
                    NSMutableDictionary *infoPlist = [NSMutableDictionary dictionaryWithContentsOfFile:plistFullPath];
                    //DLog(@"infoPlist: %@", infoPlist);
                    infoPlist[@"itemId"] = @(995367539);
                    //DLog(@"infoPlist updated: %@", infoPlist);
                    NSError *error = nil;
                    [FM removeItemAtPath:plistFullPath error:&error];
                    if (error){
                        DLog(@"error removing file: %@", plistFullPath);
                    }
                    [infoPlist writeToFile:plistFullPath atomically:true];
                }
            }
            NSString *fullPath = [tmpPath stringByAppendingPathComponent:obj.path];
            if ([ignoreFiles containsObject:obj.path.lastPathComponent]){
                DLog(@"in ignore file list, purge");
                [FM removeItemAtPath:fullPath error:nil];
            }
            
            if (self.appendedPrefix.length > 0 && [obj.path properPathComponents].count == 1) {
                if (![obj.path containsString:self.appendedPrefix]){
                    [moveRoot addObject:obj.path];
                    DLog(@"[INFO] %@ doesnt have the prefix!", obj.path);
                }
            }
            NSArray *pathComponents = [obj.path pathComponents];
            if ([pathComponents count] > 1) {
                NSString *rootPath = pathComponents[1];
                //DLog(@"\n Checking root path: %@ for file %@\n", rootPath, obj.path);
                if ([forbiddenRoots containsObject:rootPath]) {
                    DLog(@"\n [WARNING] package file: '%@' would overwrite symbolic link at '%@'\n", obj.path, rootPath);
                    NSString *privateDir = [tmpPath stringByAppendingPathComponent:@"private"];
                    if (![FM fileExistsAtPath:privateDir]){
                        [FM createDirectoryAtPath:privateDir withIntermediateDirectories:TRUE attributes:nil error:nil];
                    }
                    //take <package_name>/[rootPath] (could be etc, var, tmp) and move to <package_name>/private/[rootPath]
                    NSString *badPath = [tmpPath stringByAppendingPathComponent:rootPath];
                    NSString *newPath = [privateDir stringByAppendingPathComponent:rootPath];
                    DLog(@"\n [INFO] Moving %@ to %@...", badPath, newPath);
                    [FM moveItemAtPath:badPath toPath:newPath error:nil];
                }
            }
        }
    }];
    [moveRoot enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *oldPath = [tmpPath stringByAppendingPathComponent:obj];
        NSString *newPath = [newRoot stringByAppendingPathComponent:obj];
        DLog(@"[INFO] moving %@ to %@", oldPath, newPath);
        [FM moveItemAtPath:oldPath toPath:newPath error:nil];
    }];
    NSString *fullName = [self fullPackageName];
    if ([FM fileExistsAtPath:fullName]){
        NSString *overwritePrompt = [NSString stringWithFormat:@"%@ already exists, continuing will overwrite it.", fullName];
        BOOL shouldOverwrite = [HelperClass shouldContinueWithPrompt:overwritePrompt];
        if (shouldOverwrite) {
            [FM removeItemAtPath:fullName error:nil];
        } else {
            DLog(@"Not overwriting file... bailing out!");
            return;
        }
    }
    NSString *depArchiveInfo = [NSString stringWithFormat:@"%@ -b %@ %@", dpkg, self.packageName, fullName];
    if (fakeRoot) {
        depArchiveInfo = [NSString stringWithFormat:@"%@ %@ -b %@ %@", fakeRoot, dpkg, self.packageName, fullName];
    }
    [[HelperClass returnForProcess:depArchiveInfo] componentsJoinedByString:@"\n"];
    DLog(@"\nDone!\n\n");
}

- (NSString *)excludeContents {
    return @"path-exclude=*/share/info/*\npath-exclude=*/share/doc/*\npath-exclude=*/share/man/*";
}

- (void)repackageInCurrentDirectoryWithArch:(NSString *)newArch {
    
    NSString *fakeRoot = [HelperClass singleLineReturnForProcess:@"/usr/bin/which fakeroot"];
    NSString *pwd = [HelperClass singleLineReturnForProcess:@"/bin/pwd"];
    NSString *dpkg = [@"dpkg" runPathForSearchPath];
    DLog(@"\nProcessing file: %@\n", self.path);
    InputPackage *output = self;
    DLog(@"\nFound package: '%@' at version: '%@'...\n", output.packageName, output.version );
    NSString *tmpPath = [pwd stringByAppendingPathComponent:output.packageName];
    NSString *debian = [tmpPath stringByAppendingPathComponent:@"DEBIAN"];
    NSString *newRoot = nil;
    NSMutableArray *moveRoot = [NSMutableArray new];
    if (self.appendedPrefix) {
        newRoot = [tmpPath stringByAppendingPathComponent:self.appendedPrefix];
        [FM createDirectoryAtPath:newRoot withIntermediateDirectories:true attributes:nil error:nil];
    }
    [FM createDirectoryAtPath:tmpPath withIntermediateDirectories:TRUE attributes:nil error:nil];
    DLog(@"\nExtracting package contents for processing...\n");
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -x %@ %@", dpkg, self.path, tmpPath]];
    [FM createDirectoryAtPath:debian withIntermediateDirectories:TRUE attributes:nil error:nil];
    DLog(@"\nExtracting DEBIAN files for processing...\n");
    [HelperClass returnForProcess:[NSString stringWithFormat:@"%@ -e %@ %@", dpkg, self.path, debian]];
    NSArray *excludeArray = [self.excludeFile excludeArrayFromFile];
    if (excludeArray) {
        DLog(@"excludeARray: %@", excludeArray);
        [excludeArray enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if ([obj length] > 0){
                NSString *commandToRun = [NSString stringWithFormat:@"/bin/rm %@/%@", tmpPath, obj];
                //DLog(@"%@", commandToRun);
                NSString *deets = [HelperClass singleLineReturnForProcess:commandToRun];
                DLog(@"returned: %@", deets);
            }
            //[HelperClass returnForProcess:[NSString stringWithFormat:@"/bin/rm %@/%@", tmpPath, obj]];
        }];
    }
    //clean up any calls to uicache
    NSString *postinst = [debian stringByAppendingPathComponent:@"postinst"];
    //DLog(@"post inst: %@", postinst);
    if ([FM fileExistsAtPath:postinst]){
        NSString *postinstSource = [NSString stringWithContentsOfFile:postinst encoding:NSUTF8StringEncoding error:nil];
        if (postinstSource.length > 0){
            if ([postinstSource rangeOfString:@"uicache"].location != NSNotFound){
                NSMutableArray *lines = [[postinstSource componentsSeparatedByString:@"\n"] mutableCopy];
                //DLog(@"lines: %@", lines);
                __block NSInteger markedForDeath = NSNotFound;
                [lines enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                    //  DLog(@"obj: %@", obj);
                    if ([obj rangeOfString:@"uicache"].location != NSNotFound){
                        
                        if ([obj rangeOfString:@"echo"].location == NSNotFound){
                            DLog(@"\n");
                            DLog(@"found uicache instance '%@' on line: %lu", obj, idx);
                            markedForDeath = idx;
                        }
                        *stop = TRUE;
                    }
                }];
                if (markedForDeath != NSNotFound){
                    [lines removeObjectAtIndex:markedForDeath];
                    NSString *newString = [lines componentsJoinedByString:@"\n"];
                    [newString writeToFile:postinst atomically:TRUE];
#pragma clang diagnostic pop
                }
            }
        }
    }
    if (newArch != nil) {
        NSString *controlPath = [debian stringByAppendingPathComponent:@"control"];
        NSMutableString *controlFile = [[NSMutableString alloc] initWithContentsOfFile:controlPath encoding:NSASCIIStringEncoding error:nil];
        //@"appletvos-arm64"
        //this is hacky but should be fine, only one of the two should exist so only one will get overwritten.
        
        NSString *findString = @"iphoneos-arm";
        if (![controlFile containsString:findString]) {
            findString = @"appletvos-arm64";
        }
        
        [controlFile replaceOccurrencesOfString:findString withString:newArch options:NSLiteralSearch range:NSMakeRange(0, [controlFile length])];
        [controlFile replaceOccurrencesOfString:@"darwin-arm64" withString:newArch options:NSLiteralSearch range:NSMakeRange(0, [controlFile length])];
        
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
        [controlFile writeToFile:controlPath atomically:TRUE];
#pragma clang diagnostic pop
    }
    //[self sanitizeScriptsIfNecessary:debian];
    //at this point we have the files extracted, time to determine what needs to be changed
    NSArray *ignoreFiles = @[@".fauxsu", @".DS_Store"];
    NSArray *forbiddenRoots = @[@"etc", @"var", @"tmp"];
    [self flattenInPath:tmpPath];
    //__block NSMutableArray *_overwriteArray = [NSMutableArray new];
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        if (obj.type == BSPackageFileTypeFile || obj.type == BSPackageFileTypeDirectory){
            //DLog(@"processing path: %@", obj.path);
            if ([obj.path isEqualToString:@"/private/var/mobile/Applications/"]){
                
                NSString *badPath = [tmpPath stringByAppendingPathComponent:obj.path];
                NSString *newPath = [tmpPath stringByAppendingPathComponent:@"Applications"];
                DLog(@"\n [INFO] Moving %@ to %@...", badPath, newPath);
                [FM moveItemAtPath:badPath toPath:newPath error:nil];
                [FM removeItemAtPath:[tmpPath stringByAppendingPathComponent:@"private"] error:nil];
                *stop = TRUE;
            }
            NSString *fullPath = [tmpPath stringByAppendingPathComponent:obj.path];
            if ([ignoreFiles containsObject:obj.path.lastPathComponent]) {
                DLog(@"in ignore file list, purge");
                [FM removeItemAtPath:fullPath error:nil];
            }
            if (self.appendedPrefix.length > 0) {
                if (![obj.path containsString:self.appendedPrefix]) {
                    DLog(@"[INFO] %@ doesnt have the prefix!", obj.path);
                    [moveRoot addObject:obj.path];
                }
            }
            NSArray *pathComponents = [obj.path pathComponents];
            if ([pathComponents count] > 1) {
                NSString *rootPath = [pathComponents objectAtIndex:1];
                //DLog(@"\n Checking root path: %@ for file %@\n", rootPath, obj.path);
                if ([forbiddenRoots containsObject:rootPath]) {
                    DLog(@"\n [WARNING] package file: '%@' would overwrite symbolic link at '%@'\n", obj.path, rootPath);
                    NSString *privateDir = [tmpPath stringByAppendingPathComponent:@"private"];
                    if (![FM fileExistsAtPath:privateDir]) {
                        [FM createDirectoryAtPath:privateDir withIntermediateDirectories:TRUE attributes:nil error:nil];
                    }
                    //take <package_name>/[rootPath] (could be etc, var, tmp) and move to <package_name>/private/[rootPath]
                    NSString *badPath = [tmpPath stringByAppendingPathComponent:rootPath];
                    NSString *newPath = [privateDir stringByAppendingPathComponent:rootPath];
                    DLog(@"\n [INFO] Moving %@ to %@...", badPath, newPath);
                    [FM moveItemAtPath:badPath toPath:newPath error:nil];
                }
            }
        }
    }];
    [moveRoot enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *oldPath = [tmpPath stringByAppendingPathComponent:obj];
        NSString *newPath = [newRoot stringByAppendingPathComponent:obj];
        DLog(@"[INFO] moving %@ to %@", oldPath, newPath);
        [FM moveItemAtPath:oldPath toPath:newPath error:nil];
    }];
    
    NSString *fullName = [self fullPackageName];
    if ([FM fileExistsAtPath:fullName]){
        NSString *overwritePrompt = [NSString stringWithFormat:@"%@ already exists, do you want to overwrite it? [y/n]", fullName];
        BOOL shouldOverwrite = [HelperClass shouldContinueWithPrompt:overwritePrompt];
        if (shouldOverwrite) {
            [FM removeItemAtPath:fullName error:nil];
        } else {
            DLog(@"Not overwriting file... bailing out!");
            return;
        }
    }
    NSString *depArchiveInfo = [NSString stringWithFormat:@"%@ -b %@ %@", dpkg, self.packageName, fullName];
    if (fakeRoot) {
        depArchiveInfo = [NSString stringWithFormat:@"%@ %@ -b %@ %@", fakeRoot, dpkg, self.packageName, fullName];
    }
    [[HelperClass returnForProcess:depArchiveInfo] componentsJoinedByString:@"\n"];
    DLog(@"\nDone!\n\n");
}

- (ErrorReturn *)errorReturnForBootstrap:(NSString *)bootstrapPath {
    NSArray *ignoreFiles = @[@".fauxsu", @".DS_Store"];
    NSArray *forbiddenRoots = @[@"etc", @"var", @"tmp"];
    __block NSInteger returnValue = 0; //0 = good to go 1 = over write warning, 2 = no go
    __block NSMutableArray *_overwriteArray = [NSMutableArray new];
    [self.files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        if ([obj.fileType isEqualToString:@"file"]){
            NSString *fullPath = [bootstrapPath stringByAppendingPathComponent:obj.path];
            if ([FM fileExistsAtPath:fullPath] && ![ignoreFiles containsObject:obj.basename]){
                //DLog(@"[WARNING] overwriting a file that already exists and isn't DS_Store or .fauxsu: %@", fullPath);
                [_overwriteArray addObject:obj.path];
                //*stop = TRUE;//return FALSE;
                returnValue = 1;
            }
            NSArray *pathComponents = [obj.path pathComponents];
            if ([pathComponents count] > 1) {
                NSString *rootPath = [pathComponents objectAtIndex:1];
                if ([forbiddenRoots containsObject:rootPath]) {
                    DLog(@"\n [ERROR] package file: '%@' would overwrite symbolic link at '%@'! Exiting!\n\n", obj.path, rootPath);
                    *stop = TRUE;
                    returnValue = 2;
                }
            }
        }
    }];
    ErrorReturn *er = [ErrorReturn new];
    er.returnStatus = returnValue;
    er.overwriteFiles = _overwriteArray;
    return er;
}

@end
