
#import "InputPackageFile.h"

@implementation InputPackageFile

- (void)_setFileTypeFromRaw:(NSString *)rawType {
    
    _fileType = [InputPackageFile readableFileTypeForRawMode:rawType];
    _type = [InputPackageFile fileTypeForRawMode:rawType];
}

/*
 - denotes a regular file
 d denotes a directory
 b denotes a block special file
 c denotes a character special file
 l denotes a symbolic link
 p denotes a named pipe
 s denotes a domain socket
 */

+ (NSString* )readableFileTypeForRawMode:(NSString *)fileTypeChar {
    NSString *fileType = nil;
    if ([fileTypeChar isEqualToString:@"-"])
    { fileType = @"file"; }
    else if ([fileTypeChar isEqualToString:@"d"])
    { fileType = @"directory"; }
    else if ([fileTypeChar isEqualToString:@"b"])
    { fileType = @"block"; }
    else if ([fileTypeChar isEqualToString:@"c"])
    { fileType = @"character"; }
    else if ([fileTypeChar isEqualToString:@"l"])
    { fileType = @"link"; }
    else if ([fileTypeChar isEqualToString:@"p"])
    { fileType = @"pipe"; }
    else if ([fileTypeChar isEqualToString:@"s"])
    { fileType = @"socket"; }
    return fileType;
}

/*
 FOUNDATION_EXPORT NSFileAttributeType const NSFileTypeSocket;
 FOUNDATION_EXPORT NSFileAttributeType const NSFileTypeUnknown;
 */

+ (BSPackageFileType)fileTypeFromAttributedFileType:(NSString *)fileType {
    BSPackageFileType type = BSPackageFileTypeUnknown;
    if ([fileType isEqualToString:NSFileTypeRegular])
    { type = BSPackageFileTypeFile; }
    else if ([fileType isEqualToString:NSFileTypeDirectory])
    { type = BSPackageFileTypeDirectory; }
    else if ([fileType isEqualToString:NSFileTypeBlockSpecial])
    { type = BSPackageFileTypeBlock; }
    else if ([fileType isEqualToString:NSFileTypeCharacterSpecial])
    { type = BSPackageFileTypeCharacter; }
    else if ([fileType isEqualToString:NSFileTypeSymbolicLink])
    { type = BSPackageFileTypeLink; }
    else if ([fileType isEqualToString:NSFileTypeSocket])
    { type = BSPackageFileTypeSocket; }
    return type;
}

+ (BSPackageFileType)fileTypeForRawMode:(NSString *)fileTypeChar {
    BSPackageFileType type = BSPackageFileTypeUnknown;
    if ([fileTypeChar isEqualToString:@"-"])
    { type = BSPackageFileTypeFile; }
    else if ([fileTypeChar isEqualToString:@"d"])
    { type = BSPackageFileTypeDirectory; }
    else if ([fileTypeChar isEqualToString:@"b"])
    { type = BSPackageFileTypeBlock; }
    else if ([fileTypeChar isEqualToString:@"c"])
    { type = BSPackageFileTypeCharacter; }
    else if ([fileTypeChar isEqualToString:@"l"])
    { type = BSPackageFileTypeLink; }
    else if ([fileTypeChar isEqualToString:@"p"])
    { type = BSPackageFileTypePipe; }
    else if ([fileTypeChar isEqualToString:@"s"])
    { type = BSPackageFileTypeSocket; }
    
    return type;
    
}

/*
 NSFileCreationDate = "2023-09-20 07:09:17 +0000";
  NSFileExtensionHidden = 0;
  NSFileGroupOwnerAccountID = 0;
  NSFileGroupOwnerAccountName = wheel;
  NSFileHFSCreatorCode = 0;
  NSFileHFSTypeCode = 0;
  NSFileModificationDate = "2023-09-20 07:09:17 +0000";
  NSFileOwnerAccountID = 0;
  NSFileOwnerAccountName = root;
  NSFilePosixPermissions = 493;
  NSFileProtectionKey = NSFileProtectionCompleteUntilFirstUserAuthentication;
  NSFileReferenceCount = 1;
  NSFileSize = 68256;
  NSFileSystemFileNumber = 116616970;
  NSFileSystemNumber = 16777223;
  NSFileType = NSFileTypeRegular;
 */

- (InputPackageFile *)initWithFile:(NSString *)file inBootstrap:(NSString *)bootstrap {
    if ([file isEqualToString:@"."] || [file isEqualToString:@"./"] || file.length == 0){
        return nil;
    }
    InputPackageFile *ip = [InputPackageFile new];
    NSString *fullPath = [bootstrap stringByAppendingPathComponent:file];
    NSDictionary *attrs = [FM attributesOfItemAtPath:fullPath error:nil];
    ip.path = file;
    ip.basename = [file lastPathComponent];
    ip.owner = [NSString stringWithFormat:@"%@:%@",attrs[NSFileOwnerAccountID], attrs[NSFileGroupOwnerAccountID]];
    ip.permissions = attrs[NSFilePosixPermissions];
    ip.type = [InputPackageFile fileTypeFromAttributedFileType:attrs[NSFileType]];
    if (ip.type == BSPackageFileTypeLink) {
        ip.linkDestination = [FM destinationOfSymbolicLinkAtPath:ip.path error:nil];
    }
    return ip;
}

- (NSString*) description {
    NSString *orig = [super description];
    NSMutableDictionary *details = [NSMutableDictionary new];
    NSArray *props = [self properties];
    [props enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *cv = [self valueForKey:obj];
        if (cv){
            details[obj] = cv;
        }
    }];
    return [NSString stringWithFormat:@"%@ = %@", orig, details];
}

@end
