//
//  main.m
//  octalConversion
//
//  Created by Kevin Bradley on 6/28/18.
//  Copyright © 2018 nito. All rights reserved.
//
#import <Foundation/Foundation.h>
#include <stdio.h>
#include <errno.h>
#include <libgen.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#import <sys/utsname.h>
#import "HelperClass.h"

#define OPTION_FLAGS "o:i:ld:hcrbp:su:yn:a:Ce:fL:U"

char *progname;
char *path;

static struct option longopts[] = {
    { "octal",                     required_argument,      NULL,   'o' },
    { "input",                     required_argument,      NULL,   'i' },
    { "list",                      no_argument,            NULL,   'l' },
    { "delete",                    required_argument,      NULL,   'd' },
    { "help",                      no_argument,            NULL,   'h' },
    { "clean",                     no_argument,            NULL,   'c' },
    { "repackage",                 no_argument,            NULL,   'r' },
    { "bump",                      no_argument,            NULL,   'b' },
    { "prefix",                    required_argument,      NULL,   'p' },
    { "skip",                      no_argument,            NULL,   's' },
    { "unique",                    required_argument,      NULL,   'u' },
    { "yes",                       no_argument,            NULL,   'y' },
    { "new",                       required_argument,      NULL,   'n' },
    { "arch",                      required_argument,      NULL,   'a' },
    { "create",                    no_argument,            NULL,   'C' },
    { "exclude",                   required_argument,      NULL,   'e' },
    { "force",                     no_argument,            NULL,   'f' },
    { "contents",                  required_argument,      NULL,   'L' },
    { "unpack",                    no_argument,            NULL,   'U' },
    { NULL,                        0,                      NULL,    0  }
};

void cmd_help(void){
    printf("Usage: bootstrapTool [OPTIONS] BOOTSTRAP_FOLDER\n");
    printf("Makes various modifications to a raw bootstrap folder\n\n");
    
    printf("  -h, --help\t\t\tprints usage information\n");
    printf("  -n, --new\t\t\tcreates a brand new bootstrap with debs specified by this folder\n");
    printf("  -i, --input\t\t\tthe deb package to process\n");
    printf("  -d, --delete\t\t\tthe package to delete\n");
    printf("  -l, --list\t\t\tlist all the packages installed on this bootstrap\n");
    printf("  -c, --clean\t\t\tclean a bad status file from duplicate entries\n");
    printf("  -r, --repackage\t\trepackage to fix bad locations and to change architecture\n");
    printf("  -a, --arch\t\t\tonly valid for repackage, specify the architecture to use when repackaging\n");
    printf("  -b, --bump\t\t\trepackage and bump by one - number (ie 1.0-1 -> 1.0-2)\n");
    printf("  -s, --skip\t\t\tskip code signature checks when bumping and repackaging\n");
    printf("  -y, --yes\t\t\talways answer yes to prompts to overwrite files\n");
    printf("  -p, --prefix\t\t\tthe prefix to append to all packages when repackaging or bumping a version number\n");
    printf("  -C, --create\t\t\tcreate a package from the bootstrap with an interactive list\n");
    printf("  -e, --exclude\t\t\tchoose a file list for excluded files when bumping, repackaging or creating a package\n");
    printf("  -f, --force\t\t\tforce an install operation, ignoring 'forbidden' roots\n");
    printf("  -L, --contents\t\tlist the contents of an installed package, required a bootstrap path\n");
    printf("  -U, --unpack\t\tmark the package as unpacked rather than installed in the status file\n");
    printf("\n");
}

int main(int argc, char **argv) {
    @autoreleasepool {
        
        setuid(0);
        setgid(0);
        progname = basename(argv[0]);
        path = dirname(argv[0]);
        int flag;
        NSString *octalFile = nil;
        NSString *debFile = nil;
        NSString *bootstrapPath = nil;
        NSString *deletePackage = nil;
        NSString *prefix = nil;
        NSString *packagesFolder = nil;
        NSString *arch = nil;
        NSString *excludeFile = nil;
        BOOL listPackage = FALSE;
        BOOL repackage = FALSE;
        BOOL skip = FALSE;
        BOOL bump = FALSE;
        BOOL clean = FALSE;
        BOOL forceYes = FALSE;
        BOOL new = FALSE;
        BOOL create = FALSE;
        BOOL force = FALSE;
        BOOL unpack = FALSE;
        NSString *targetPkg = nil; //for --contents/L flag
        /*
        NSString *searchPathString = @"/Users/bradleyk/.rbenv/shims:/Users/bradleyk/Library/Python/3.9/bin:/Users/bradleyk/Library/Python/3.8/bin:/usr/local/lib/ruby/gems/3.0.0/bin:/usr/local/lib/ruby/gems/bin:/usr/local/opt/ruby/bin:/opt/local/bin:/opt/local/sbin:/Users/bradleyk/local/bin:/usr/local/mysql/bin:/opt/local/bin:/opt/local/sbin:/usr/bin:/Users/bradleyk/Projects/goprojects/bin:/Users/bradleyk/Projects/theos/bin:/usr/local/bin/:/usr/local/bin:/bin/:/sbin/:/usr/sbin/";
        NSString *target = @"dpkg-deb";
        NSString *gotIm = [target pathForBin:target searchPaths:searchPathString];
        DLog(@"gotIm: %@", gotIm);
        return 0;
         */
        NSInteger loadAttempt = 0;
        BOOL isLoaded = false;
        do {
            isLoaded = true;//[[$LSApplicationWorkspace defaultWorkspace] registerApplicationDictionary:dict];
            if (!isLoaded) {
                loadAttempt++;
                usleep(1.0*USEC_PER_SEC);
                
            }
        } while (isLoaded == false);
        NSString *up = nil;
        while ((flag = getopt_long(argc, argv, OPTION_FLAGS, longopts, NULL)) != -1) {
            switch(flag) {
                case 'f':
                    force = TRUE;
                    break;
                case 'U':
                    unpack = TRUE;
                    break;
                case 'L':
                    targetPkg = [NSString stringWithUTF8String:optarg];
                    break;
                case 'o':
                    octalFile = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'i':
                    debFile = [NSString stringWithUTF8String:optarg];
                    break;
                case 'y':
                    forceYes = TRUE;
                    break;
                    
                case 'l':
                    listPackage = TRUE;
                    break;
                    
                case 'd':
                    deletePackage = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'b':
                    bump = TRUE;
                    break;
                    
                case 'r':
                    repackage = TRUE;
                    break;
                    
                case 'c':
                    clean = TRUE;
                    break;
                    
                case 'p':
                    prefix = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 's':
                    skip = TRUE;
                    break;
                
                case 'u':
                    up = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'a':
                    arch = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'n':
                    new = TRUE;
                    packagesFolder = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'C':
                    create = TRUE;
                    break;
                    
                case 'e':
                    excludeFile = [NSString stringWithUTF8String:optarg];
                    break;
                    
                default:
                    cmd_help();
                    return -1;
            }
        }
        if (argc-optind == 1) {
            argc -= optind;
            argv += optind;
            bootstrapPath = [NSString stringWithUTF8String:argv[0]];
            //yDLog(@"bootstrap path: %@?", bootstrapPath);
            //return 0;
        }
        if (up) {
            NSArray *contents = [FM contentsOfDirectoryAtPath:up error:nil];
            NSString *newFile = [[up lastPathComponent] stringByAppendingPathExtension:@"plist"];
            NSArray *unique = [contents uniquePackageNames];
            [unique writeToFile:newFile atomically:true];
            
            DLog(@"%@", unique);
            return 0;
        }
        
        if (targetPkg) {
            if (bootstrapPath) {
                NSString *statusFile = [bootstrapPath relativeStatusFilePath];
                NSArray *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile bootstrapPath:bootstrapPath];
                StatusPackageModel *filtered = [[installedPackages filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"package == %@", targetPkg]] firstObject];
                DLog(@"found package: %@", filtered);
                DLog(@"contents: %@", [filtered filesPlain]);
            } else {
                DLog(@"This option requires a bootstrap path!");
            }
            return 0;
        }
        
        if (bump == TRUE && debFile) {
            if (![FM fileExistsAtPath:debFile]){
                DLog(@"%@ doesnt exist!, exiting...", debFile);
                return -1;
            }
            DLog(@"skip signature checks: %d", skip);
            return [HelperClass bumpPackageAtPath:debFile withPrefix:prefix sigChecks:skip excludeFile:excludeFile];
        }
        
        if (repackage == TRUE && debFile) {
            if (![FM fileExistsAtPath:debFile]){
                DLog(@"%@ doesnt exist!, exiting...", debFile);
                return -1;
            }
            return [HelperClass repackagePackageAtPath:debFile withPrefix:prefix sigChecks:skip arch:arch excludeFile:excludeFile];
        }
        
        //DLog(@"folder: %@ project: %@", folder, project);
        if (clean && bootstrapPath) {
            return [HelperClass cleanBootstrapAtPath:bootstrapPath];
        } else if (clean) {
            DLog(@"\nYou need to choose a bootstrap path to clean!\n\n");
            return 1;
        }
        if (deletePackage && bootstrapPath) {
            return [HelperClass deletePackage:deletePackage inBootstrap:bootstrapPath];
        } else if (deletePackage) {
            DLog(@"\nYou need to choose a bootstrap path to delete from!\n\n");
            return 1;
        }
        if (octalFile) {
            if([FM fileExistsAtPath:octalFile]) {
                NSDictionary *attrs = [FM attributesOfItemAtPath:octalFile error:nil];
                NSNumber *posixPerms = attrs[NSFilePosixPermissions];
                DLog(@"perms: %@", posixPerms);
                return 0;
            }
            NSString *octal = [HelperClass octalFromSymbols:octalFile];
            DLog(@"%@", octal);
            return 0;
        }
        if (create && bootstrapPath) {
            NSString *statusFile = [bootstrapPath relativeStatusFilePath];
            NSArray *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile bootstrapPath:bootstrapPath];
            NSString *numered = [HelperClass numberedArrayFromStatusFile:statusFile bootstrapPath:bootstrapPath];
            DLog(@"Choose a package from the list below to create from the bootstrap\n");
            DLog(@"%@", numered);
            DLog(@"\n");
            NSInteger getNumber = [HelperClass getNumberFromPrompt:@"Choose a package index from above: "];
            while (getNumber > installedPackages.count) {
                getNumber = [HelperClass getNumberFromPrompt:@"Choose a package index from above: "];
            }
            DLog(@"chosen package: %lu: %@", getNumber, installedPackages[getNumber-1]);
            StatusPackageModel *model = installedPackages[getNumber-1];
            DLog(@"info files: %@", model.infoFiles);
            DLog(@"files: %@", model.files);
            [model createPackage];
            return 0;
        } else if (create) {
            
            DLog(@"\nYou need to choose a bootstrap path to create packages from!\n\n");
            return 1;
        }
        if (listPackage && bootstrapPath) {
            NSString *statusFile = [bootstrapPath relativeStatusFilePath];
            NSArray *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile bootstrapPath:bootstrapPath];
            DLog(@"%@", installedPackages);
            return 0;
            
        } else if (listPackage) {
            
            DLog(@"\nYou need to choose a bootstrap path to list packages from!\n\n");
            return 1;
        }
        
        if (new && bootstrapPath && packagesFolder) {
            return [HelperClass createNewBootstrap:bootstrapPath withPackages:packagesFolder];
        } else if (new && bootstrapPath) {
            DLog(@"you need to choose a packages folder to install from when creating a new bootstrap!");
        } else if (new) {
            DLog(@"you need to choose a packages folder to install from when creating a new bootstrap, and a bootstrap folder needs to be specified as well.");
        }
        
        if (debFile && bootstrapPath) {
            if (![FM fileExistsAtPath:debFile]){
                DLog(@"%@ doesnt exist!, exiting...", debFile);
                return -1;
            }
            DLog(@"\nProcessing file: %@\n", debFile);
            InputPackage *output = [HelperClass packageForDeb:debFile];
            output.forceOverwrite = forceYes;
            output.excludeFile = excludeFile;
            output.forceInstall = force;
            output.unpack = unpack;
            return [output installToBootstrapPath:bootstrapPath];
            
        } else if (debFile) {
            DLog(@"\nYou need to choose a bootstrap path to install the package to!\n\n");
            return 1;
        }
        //if we got this far.. nothing happened, print help.
        cmd_help();
        
    }
    return 0;
}
